@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
  *
- * Copyright (C) 2014 Matthias Klumpp <matthias@tenstral.net>
+ * Copyright (C) 2014-2016 Matthias Klumpp <matthias@tenstral.net>
  *
  * Licensed under the GNU Lesser General Public License Version 2.1
  *
@@ -36,6 +36,7 @@
 #include "li-keyring.h"
 #include "li-installer.h"
 #include "li-update-item.h"
+#include "li-package-graph.h"
 
 #include "li-dbus-interface.h"
 
@@ -295,12 +296,15 @@ li_manager_update_software_table (LiManager *mgr, GError **error)
 /**
  * li_manager_get_software_list:
  *
- * Returns: (transfer container) (element-type LiPkgInfo): A list of installed and available software
+ * Returns: (transfer full) (element-type LiPkgInfo): A list of installed and available software
  **/
-GList*
+GPtrArray*
 li_manager_get_software_list (LiManager *mgr, GError **error)
 {
+	GList *l = NULL;
+	g_autoptr(GList) list = NULL;
 	GError *tmp_error = NULL;
+	GPtrArray *res = NULL;
 	LiManagerPrivate *priv = GET_PRIVATE (mgr);
 
 	li_manager_update_software_table (mgr, &tmp_error);
@@ -309,7 +313,14 @@ li_manager_get_software_list (LiManager *mgr, GError **error)
 		return NULL;
 	}
 
-	return g_hash_table_get_values (priv->pkgs);
+	res = g_ptr_array_new_with_free_func (g_object_unref);
+	list = g_hash_table_get_values (priv->pkgs);
+	for (l = list; l != NULL; l = l->next) {
+		LiPkgInfo *pki = LI_PKG_INFO (l->data);
+		g_ptr_array_add (res, g_object_ref (pki));
+	}
+
+	return res;
 }
 
 /**
@@ -505,19 +516,20 @@ li_manager_remove_exported_files (GFile *file, GError **error)
 	g_object_unref (ir);
 
 	while (TRUE) {
-		gchar **parts;
+		g_auto(GStrv) parts = NULL;
 		line = g_data_input_stream_read_line (dis, NULL, NULL, NULL);
 		if (line == NULL) {
 			break;
 		}
 
 		parts = g_strsplit (line, "\t", 2);
-		if (parts[1] == NULL) {
-			g_strfreev (parts);
+		if (parts[1] == NULL)
 			continue;
-		}
 
 		if (g_str_has_prefix (parts[1], "/")) {
+			/* don't fail if file is already gone */
+			if (!g_file_test (parts[1], G_FILE_TEST_EXISTS))
+				continue;
 			/* delete file */
 			res = g_remove (parts[1]);
 			if (res != 0) {
@@ -525,11 +537,9 @@ li_manager_remove_exported_files (GFile *file, GError **error)
 					LI_MANAGER_ERROR,
 					LI_MANAGER_ERROR_REMOVE_FAILED,
 					_("Could not delete file '%s'"), parts[1]);
-				g_strfreev (parts);
 				goto out;
 			}
 		}
-		g_strfreev (parts);
 	}
 
 out:
@@ -717,8 +727,8 @@ li_manager_remove_software (LiManager *mgr, const gchar *pkgid, GError **error)
 	g_ptr_array_add (pkg_array, pki);
 	rt = li_manager_find_runtime_with_members (mgr, pkg_array);
 	if (rt != NULL) {
-		g_autoptr(GList) sw = NULL;
-		GList *l;
+		g_autoptr(GPtrArray) sw = NULL;
+		guint i;
 		gboolean dependency_found = FALSE;
 
 		/* check if this software is in use somewhere */
@@ -728,8 +738,8 @@ li_manager_remove_software (LiManager *mgr, const gchar *pkgid, GError **error)
 			return FALSE;
 		}
 
-		for (l = sw; l != NULL; l = l->next) {
-			LiPkgInfo *pki2 = LI_PKG_INFO (l->data);
+		for (i = 0; i < sw->len; i++) {
+			LiPkgInfo *pki2 = LI_PKG_INFO (g_ptr_array_index (sw, i));
 
 			/* check if the package is actually installed */
 			if (!li_pkg_info_has_flag (pki2, LI_PACKAGE_FLAG_INSTALLED))
@@ -904,15 +914,24 @@ out:
  * li_manager_cleanup:
  *
  * Remove unnecessary software.
+ * Limba automatically determines whether software is still needed.
+ * Only software which has been explicitly installed is kept, anything
+ * else will be automatically cleaned as soon as nothing needs it anymore.
+ * The cleanup routine will remove faded bundles, so it is recommended to
+ * runt it when to Limba-installed app is still in use.
+ * By default, cleanup happens when the system is shutting down.
  */
 gboolean
 li_manager_cleanup (LiManager *mgr, GError **error)
 {
 	g_autoptr(GHashTable) sws = NULL;
-	g_autoptr(GHashTable) rts = NULL;
+	g_autoptr(GHashTable) remove_rts = NULL;
+	g_autoptr(GPtrArray) active_rts = NULL;
 	g_autoptr(GList) sw_list = NULL;
 	g_autoptr(GList) list = NULL;
-	GPtrArray *rt_array;
+	g_autoptr(LiPackageGraph) pg = NULL;
+	g_autoptr(GPtrArray) removal_candidates = NULL;
+	GPtrArray *all_rts_array;
 	guint i;
 	GList *l;
 	GError *tmp_error = NULL;
@@ -926,9 +945,6 @@ li_manager_cleanup (LiManager *mgr, GError **error)
 		goto out;
 	}
 
-	rts = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_object_unref);
-	sws = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_object_unref);
-
 	/* load software list */
 	sws = li_manager_get_installed_software (mgr, &tmp_error);
 	if (tmp_error != NULL) {
@@ -958,52 +974,111 @@ li_manager_cleanup (LiManager *mgr, GError **error)
 		/* installed software might have changed */
 		li_manager_reset_cached_data (mgr);
 
-		g_list_free (sw_list);
 		g_hash_table_unref (sws);
 		sws = li_manager_get_installed_software (mgr, &tmp_error);
 		if (tmp_error != NULL) {
 			g_propagate_error (error, tmp_error);
 			return FALSE;
 		}
-
+		g_list_free (sw_list);
 		sw_list = g_hash_table_get_values (sws);
 	}
 
-	/* remove every software from the list which is member of a runtime */
-	rt_array = li_manager_get_installed_runtimes (mgr);
-	for (i = 0; i < rt_array->len; i++) {
+	/* convert removal GList into GPtrArray */
+	removal_candidates = g_ptr_array_new ();
+	for (l = sw_list; l != NULL; l = l->next) {
+		LiPkgInfo *pki = LI_PKG_INFO (l->data);
+		g_ptr_array_add (removal_candidates, pki);
+	}
+
+	/* create a dependency graph of the remaining packages, we will need it later */
+	pg = li_package_graph_new_from_pkiarray (removal_candidates, &tmp_error);
+	if (tmp_error != NULL) {
+		g_propagate_error (error, tmp_error);
+		goto out;
+	}
+
+	/* load runtime list */
+	all_rts_array = li_manager_get_installed_runtimes (mgr);
+	remove_rts = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
+
+	/* create a map of all runtimes we have */
+	for (i = 0; i < all_rts_array->len; i++) {
+		LiRuntime *rt = LI_RUNTIME (g_ptr_array_index (all_rts_array, i));
+		g_hash_table_insert (remove_rts,
+					g_strdup (li_runtime_get_uuid (rt)),
+					rt);
+	}
+
+	/* remove every software from the kill-list which was installed manually,
+	 * and drop runtimes from the kill-list which are in use */
+	for (l = sw_list; l != NULL; l = l->next) {
+		const gchar *rt_uuid = NULL;
+		LiPkgInfo *pki = LI_PKG_INFO (l->data);
+
+		rt_uuid = li_pkg_info_get_runtime_dependency (pki);
+		if (rt_uuid != NULL) {
+			/* remove this runtime from the kill-list, it is in use */
+			g_hash_table_remove (remove_rts, rt_uuid);
+		}
+
+		if (!li_pkg_info_has_flag (pki, LI_PACKAGE_FLAG_AUTOMATIC)) {
+			/* manual package */
+			g_debug ("Manual: %s", li_pkg_info_get_id (pki));
+			g_hash_table_remove (sws, li_pkg_info_get_id (pki));
+			continue;
+		}
+	}
+
+	/* drop unused runtimes now and create list of still valid runtimes */
+	active_rts = g_ptr_array_new_with_free_func (g_object_unref);
+	for (i = 0; i < all_rts_array->len; i++) {
+		const gchar *rt_uuid = NULL;
+		LiRuntime *rt = LI_RUNTIME (g_ptr_array_index (all_rts_array, i));
+
+		rt_uuid = li_runtime_get_uuid (rt);
+		if (g_hash_table_contains (remove_rts, rt_uuid)) {
+			/* runtime is in the kill-list, drop it */
+			li_runtime_remove (rt);
+		} else {
+			g_ptr_array_add (active_rts, g_object_ref (rt));
+		}
+	}
+
+	/* remove every software from the kill-list which is member of a runtime */
+	for (i = 0; i < active_rts->len; i++) {
 		gchar **rt_members;
 		guint len;
 		guint j;
-		LiRuntime *rt = LI_RUNTIME (g_ptr_array_index (rt_array, i));
+		LiRuntime *rt = LI_RUNTIME (g_ptr_array_index (active_rts, i));
 
 		rt_members = (gchar**) g_hash_table_get_keys_as_array (li_runtime_get_members (rt), &len);
 		for (j = 0; j < len; j++) {
+			g_debug ("Inuse: %s", rt_members[j]);
 			g_hash_table_remove (sws, rt_members[j]);
 		}
 		g_free (rt_members);
-
-		g_hash_table_insert (rts,
-					g_strdup (li_runtime_get_uuid (rt)),
-					g_object_ref (rt));
 	}
 
-	/* remove every software from the kill-list which references a valid runtime */
 	list = g_hash_table_get_values (sws);
-	for (l = list; l != NULL; l = l->next) {
+	for (l = list; l != NULL; l= l->next) {
+		const gchar *pkid;
+
 		LiPkgInfo *pki = LI_PKG_INFO (l->data);
+		pkid = li_pkg_info_get_id (pki);
 
-		if (g_hash_table_lookup (rts, li_pkg_info_get_runtime_dependency (pki)) != NULL) {
-			g_hash_table_remove (sws, li_pkg_info_get_id (pki));
+		if (li_package_graph_node_get_any_parent_manual (pg, pki)) {
+			g_debug ("Skip: %s (dependency of non-auto)", pkid);
+			continue;
 		}
-	}
-	g_list_free (list);
 
-	list = g_hash_table_get_values (sws);
-	for (l = list; l != NULL; l = l->next) {
-		li_manager_remove_software (mgr, li_pkg_info_get_id (LI_PKG_INFO (l->data)), error);
-		if (error != NULL)
+		li_manager_remove_software (mgr,
+					    pkid,
+					    &tmp_error);
+		if (tmp_error != NULL) {
+			g_propagate_error (error, tmp_error);
 			goto out;
+		}
 	}
 
 	/* cleanup tmp dir */
@@ -1025,6 +1100,7 @@ out:
 
 /**
  * li_manager_refresh_cache:
+ * @mgr: An instance of #LiManager
  *
  * Refresh the cache of available packages.
  */
@@ -1049,17 +1125,52 @@ li_manager_refresh_cache (LiManager *mgr, GError **error)
 }
 
 /**
- * li_manager_receive_key:
+ * li_manager_refresh_keyring:
+ * @mgr: An instance of #LiManager
  *
- * Download a PGP key and add it to the database of highly trusted keys.
+ * Look for new trusted keys and update the keyring.
  */
 void
-li_manager_receive_key (LiManager *mgr, const gchar *fpr, GError **error)
+li_manager_refresh_keyring (LiManager *mgr, GError **error)
 {
 	LiKeyring *kr;
 
 	kr = li_keyring_new ();
-	li_keyring_import_key (kr, fpr, LI_KEYRING_KIND_USER, error);
+	li_keyring_refresh_keys (kr, error);
+	g_object_unref (kr);
+}
+
+/**
+ * li_manager_trust_key:
+ * @mgr: An instance of #LiManager
+ * @fpr: The fingerprint of the key to add.
+ *
+ * Download a PGP key and add it to the database of trusted keys.
+ */
+void
+li_manager_trust_key (LiManager *mgr, const gchar *fpr, GError **error)
+{
+	LiKeyring *kr;
+
+	kr = li_keyring_new ();
+	li_keyring_add_key (kr, fpr, error);
+	g_object_unref (kr);
+}
+
+/**
+ * li_manager_trust_key_file:
+ * @mgr: An instance of #LiManager
+ * @fname: The filename of the GPG key that should be added.
+ *
+ * Download a PGP key and add it to the database of trusted keys.
+ */
+void
+li_manager_trust_key_file (LiManager *mgr, const gchar *fname, GError **error)
+{
+	LiKeyring *kr;
+
+	kr = li_keyring_new ();
+	li_keyring_add_key_file (kr, fname, error);
 	g_object_unref (kr);
 }
 
