@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
  *
- * Copyright (C) 2014 Matthias Klumpp <matthias@tenstral.net>
+ * Copyright (C) 2014-2016 Matthias Klumpp <matthias@tenstral.net>
  *
  * Licensed under the GNU Lesser General Public License Version 2.1
  *
@@ -58,7 +58,7 @@ struct _LiPackagePrivate
 
 	gchar *install_root;
 	gchar *id;
-	GList *embedded_packages;
+	GPtrArray *embedded_packages;
 
 	LiKeyring *kr;
 	gchar *signature_data;
@@ -100,7 +100,7 @@ li_package_finalize (GObject *object)
 	if (priv->cpt != NULL)
 		g_object_unref (priv->cpt);
 	if (priv->embedded_packages != NULL)
-		g_list_free_full (priv->embedded_packages, g_object_unref);
+		g_ptr_array_unref (priv->embedded_packages);
 	if (priv->tmp_dir != NULL) {
 		li_delete_dir_recursive (priv->tmp_dir);
 		g_free (priv->tmp_dir);
@@ -150,7 +150,7 @@ li_package_emit_progress (LiPackage *pkg)
 	guint percentage;
 	LiPackagePrivate *priv = GET_PRIVATE (pkg);
 
-	percentage = round (100 / (double) priv->max_progress * priv->progress);
+	percentage = round ((100 / (double) priv->max_progress) * priv->progress);
 	g_signal_emit (pkg, signals[SIGNAL_PROGRESS], 0,
 					percentage);
 }
@@ -185,7 +185,7 @@ li_package_cache_progress_cb (LiPkgCache *cache, guint cache_percentage, const g
 	if (g_strcmp0 (priv->id, id) != 0)
 		return;
 
-	percentage = round (100 / (double) priv->max_progress * (priv->progress+cache_percentage));
+	percentage = round ((100 / (double) priv->max_progress) * (priv->progress+cache_percentage));
 
 	g_signal_emit (pkg, signals[SIGNAL_PROGRESS], 0,
 					percentage);
@@ -218,7 +218,6 @@ li_package_extract_entry_to (LiPackage *pkg, struct archive* ar, struct archive_
 {
 	g_autofree gchar *fname = NULL;
 	const gchar *cstr;
-	const gchar *link_target;
 	gchar *str;
 	gboolean ret;
 	gint res;
@@ -234,7 +233,7 @@ li_package_extract_entry_to (LiPackage *pkg, struct archive* ar, struct archive_
 
 	filetype = archive_entry_filetype (e);
 	if (filetype == S_IFDIR) {
-		/* we don't extract directories */
+		/* we don't extract directories explicitly */
 		return TRUE;
 	}
 
@@ -252,16 +251,19 @@ li_package_extract_entry_to (LiPackage *pkg, struct archive* ar, struct archive_
 	}
 
 	/* check if we are dealing with a symlink */
-	link_target = archive_entry_symlink (e);
-	if ((filetype == S_IFLNK) && (link_target != NULL)) {
-		res = symlink (link_target, fname);
-	} else {
-		link_target = archive_entry_hardlink (e);
-		if (link_target != NULL) {
-			res = symlink (link_target, fname);
+	if (filetype == S_IFLNK) {
+		const gchar *link_target = NULL;
+		link_target = archive_entry_symlink (e);
+
+		if (link_target == NULL) {
+			g_set_error (error,
+				LI_PACKAGE_ERROR,
+				LI_PACKAGE_ERROR_EXTRACT,
+				_("Unable to read symlink destination for file: %s"), fname);
+			return FALSE;
 		}
-	}
-	if (link_target != NULL) {
+
+		res = symlink (link_target, fname);
 		if (res != 0) {
 			g_set_error (error,
 				LI_PACKAGE_ERROR,
@@ -274,8 +276,8 @@ li_package_extract_entry_to (LiPackage *pkg, struct archive* ar, struct archive_
 	}
 
 	if (filetype != S_IFREG) {
-		/* do we really have to extract every type of file?
-		 * Symlinks and regular files should be enough for now */
+		/* Extracting symlinks and regular files should be enough for now,
+		 * to prevent issues by creating (unwanted?) special files. */
 		g_debug ("Skipped extraction of file '%s': No regular file.",
 				 archive_entry_pathname (e));
 		return TRUE;
@@ -328,14 +330,16 @@ li_package_extract_entry_to (LiPackage *pkg, struct archive* ar, struct archive_
 	}
 
 done:
-	/* apply permissions from the archive */
-	res = chmod (fname, archive_entry_mode (e));
-	if (res != 0) {
-		g_set_error (error,
-				LI_PACKAGE_ERROR,
-				LI_PACKAGE_ERROR_FAILED,
-				_("Unable to set permissions on file '%s'. Error: %s"), fname, g_strerror (errno));
-		return FALSE;
+	if (filetype != S_IFLNK) {
+		/* ensure correct permissions are applied */
+		res = chmod (fname, archive_entry_mode (e));
+		if (res != 0) {
+			g_set_error (error,
+					LI_PACKAGE_ERROR,
+					LI_PACKAGE_ERROR_FAILED,
+					_("Unable to set permissions on file '%s'. Error: %s"), fname, g_strerror (errno));
+			return FALSE;
+		}
 	}
 
 	return ret;
@@ -421,20 +425,13 @@ li_package_read_component_data (LiPackage *pkg, const gchar *data, GError **erro
 		return FALSE;
 	}
 
-	tmp = li_str_replace (as_component_get_id (priv->cpt), ".desktop", "");
-	g_strstrip (tmp);
-	if ((tmp == NULL) || (g_strcmp0 (tmp, "") == 0)) {
-		g_free (tmp);
-		tmp = li_str_replace (as_component_get_name (priv->cpt), " ", "_");
-		if ((tmp == NULL) || (g_strcmp0 (tmp, "") == 0)) {
-			g_free (tmp);
-			/* no package name is found, we give up */
-			g_set_error (error,
+	tmp = li_get_pkgname_from_component (priv->cpt);
+	if (tmp == NULL) {
+		g_set_error (error,
 				LI_PACKAGE_ERROR,
 				LI_PACKAGE_ERROR_DATA_MISSING,
 				_("Could not determine package name."));
-			return FALSE;
-		}
+		return FALSE;
 	}
 	li_pkg_info_set_name (priv->info, tmp);
 	g_free (tmp);
@@ -470,8 +467,6 @@ li_package_open_file (LiPackage *pkg, const gchar *filename, GError **error)
 	gchar *tmp_str;
 	GError *tmp_error = NULL;
 	gboolean ret = FALSE;
-	GPtrArray *epkgs;
-	guint i;
 	LiPackagePrivate *priv = GET_PRIVATE (pkg);
 
 	if (!g_file_test (filename, G_FILE_TEST_IS_REGULAR)) {
@@ -555,18 +550,10 @@ li_package_open_file (LiPackage *pkg, const gchar *filename, GError **error)
 						g_compute_checksum_for_string (G_CHECKSUM_SHA256, index_data, -1));
 			g_free (index_data);
 
-			/* clean up a possible previous list of packages */
-			if (priv->embedded_packages != NULL) {
-				g_list_free_full (priv->embedded_packages, g_object_unref);
-				priv->embedded_packages = NULL;
-			}
-
-			/* convert GPtrArray to GList */
-			epkgs = li_pkg_index_get_packages (idx);
-			for (i = 0; i < epkgs->len; i++) {
-				LiPkgInfo *pki = LI_PKG_INFO (g_ptr_array_index (epkgs, i));
-				priv->embedded_packages = g_list_append (priv->embedded_packages, g_object_ref (pki));
-			}
+			/* take ownership of the embedded packages list */
+			if (priv->embedded_packages != NULL)
+				g_ptr_array_unref (priv->embedded_packages);
+			priv->embedded_packages = g_ptr_array_ref (li_pkg_index_get_packages (idx));
 
 			g_object_unref (idx);
 		} else if (g_strcmp0 (pathname, "_signature") == 0) {
@@ -633,7 +620,7 @@ li_package_open_remote (LiPackage *pkg, LiPkgCache *cache, const gchar *pkid, GE
 
 	/* connect cache signals */
 	g_signal_connect (priv->cache, "progress",
-						G_CALLBACK (li_package_cache_progress_cb), pkg);
+				G_CALLBACK (li_package_cache_progress_cb), pkg);
 
 	priv->max_progress += 100;
 
@@ -926,10 +913,13 @@ li_package_download (LiPackage *pkg, GError **error)
 	li_package_emit_stage_change (pkg, LI_PACKAGE_STAGE_DOWNLOADING);
 
 	priv->max_progress += 100;
-	pkg_fname = li_pkg_cache_fetch_remote (priv->cache, li_pkg_info_get_id (priv->info), &tmp_error);
+	pkg_fname = li_pkg_cache_fetch_remote (priv->cache,
+					       li_pkg_info_get_id (priv->info),
+					       &tmp_error);
 	if (tmp_error != NULL) {
-		g_propagate_prefixed_error (error, tmp_error,
-								_("Unable to download package:"));
+		g_propagate_prefixed_error (error,
+					    tmp_error,
+					    _("Unable to download package:"));
 		return FALSE;
 	}
 
@@ -938,7 +928,6 @@ li_package_download (LiPackage *pkg, GError **error)
 		g_propagate_error (error, tmp_error);
 		return FALSE;
 	}
-
 	/* FIXME: Opening the file incremented our max_progress, so we decrease it again to get a 50/50 split between
 	 * download and installation. This is a bit ugly, and will have to be fixed when we emit progress for the installation
 	 * itself as well. */
@@ -1260,6 +1249,7 @@ li_package_verify_signature (LiPackage *pkg, GError **error)
 
 	level = li_keyring_process_signature (priv->kr, priv->signature_data, &sig_content, &priv->sig_fpr, &tmp_error);
 	if (tmp_error != NULL) {
+		priv->tlevel = level;
 		g_propagate_error (error, tmp_error);
 		goto out;
 	}
@@ -1395,7 +1385,7 @@ li_package_get_info (LiPackage *pkg)
  *
  * Returns: (transfer none) (element-type LiPkgInfo): Array of available embedded packages
  */
-GList*
+GPtrArray*
 li_package_get_embedded_packages (LiPackage *pkg)
 {
 	LiPackagePrivate *priv = GET_PRIVATE (pkg);
@@ -1411,7 +1401,7 @@ gboolean
 li_package_has_embedded_packages (LiPackage *pkg)
 {
 	LiPackagePrivate *priv = GET_PRIVATE (pkg);
-	return (priv->embedded_packages != NULL ) && (g_list_length (priv->embedded_packages) > 0);
+	return (priv->embedded_packages != NULL ) && (priv->embedded_packages->len > 0);
 }
 
 /**
