@@ -164,26 +164,56 @@ li_pkg_builder_write_payload (const gchar *input_dir, const gchar *out_fname, Li
 			}
 		}
 
-		stat(fname, &st);
-		entry = archive_entry_new ();
-		archive_entry_set_pathname (entry, ar_fname);
-
-		archive_entry_copy_stat (entry, &st);
-		archive_write_header (a, entry);
-
-		fd = open (fname, O_RDONLY);
-		if (fd < 0) {
-			g_warning ("Could not open file '%s' for reading. Skipping it.", fname);
-			archive_entry_free (entry);
+		if (lstat (fname, &st) != 0) {
+			g_warning ("Could not stat file '%s'. Skipping it.", fname);
 			continue;
 		}
+		entry = archive_entry_new ();
+		archive_entry_set_pathname (entry, ar_fname);
+		archive_entry_copy_stat (entry, &st);
 
-		len = read (fd, buff, sizeof (buff));
-		while (len > 0) {
-			archive_write_data (a, buff, len);
-			len = read(fd, buff, sizeof (buff));
+		/* handle symbolic links */
+		if (S_ISLNK (st.st_mode)) {
+			g_autofree gchar *linktarget = NULL;
+			ssize_t r;
+
+			linktarget = malloc (st.st_size + 1);
+			r = readlink (fname, linktarget, st.st_size + 1);
+			if (r < 0) {
+				g_warning ("Could not follow symlink '%s', readlink failed. Skipping it.", fname);
+				archive_entry_free (entry);
+				continue;
+			}
+			if (r > st.st_size) {
+				g_warning ("Could not follow symlink '%s', buffer too small. Skipping it.", fname);
+				archive_entry_free (entry);
+				continue;
+			}
+			linktarget[st.st_size] = '\0';
+
+			archive_entry_set_symlink (entry, linktarget);
 		}
-		close (fd);
+
+		/* write file header in tarball */
+		archive_write_header (a, entry);
+
+		/* add data, in case we have a regular file */
+		if (S_ISREG (st.st_mode)) {
+			fd = open (fname, O_RDONLY);
+			if (fd < 0) {
+				g_warning ("Could not open file '%s' for reading. Skipping it.", fname);
+				archive_entry_free (entry);
+				continue;
+			}
+
+			len = read (fd, buff, sizeof (buff));
+			while (len > 0) {
+				archive_write_data (a, buff, len);
+				len = read(fd, buff, sizeof (buff));
+			}
+			close (fd);
+		}
+
 		archive_entry_free (entry);
 	}
 
@@ -412,6 +442,7 @@ li_pkg_builder_sign_data (LiPkgBuilder *builder, const gchar *data, gpgme_sig_mo
 	if (priv->gpg_key != NULL) {
 		gpgme_key_t akey;
 
+		g_debug ("Explicitly using key: %s", priv->gpg_key);
 		err = gpgme_get_key (ctx, priv->gpg_key, &akey, 1);
 		if (err != 0) {
 			g_set_error (error,
@@ -435,6 +466,8 @@ li_pkg_builder_sign_data (LiPkgBuilder *builder, const gchar *data, gpgme_sig_mo
 		}
 
 		gpgme_key_unref (akey);
+	} else {
+		g_debug ("Using default key for signing.");
 	}
 
 	err = gpgme_data_new_from_mem (&din, data, strlen (data), 0);
@@ -533,7 +566,6 @@ li_pkg_builder_sign_package (LiPkgBuilder *builder, const gchar *tmp_dir, GPtrAr
 		g_propagate_error (error, tmp_error);
 		return NULL;
 	}
-	g_print ("%s\n", _("Package signed."));
 
 	sig_fname = g_build_filename (tmp_dir, "_signature", NULL);
 	file = fopen (sig_fname, "w");
@@ -595,16 +627,11 @@ li_pkg_builder_write_dsc_file (LiPkgBuilder *builder, const gchar *pkg_fname_rt,
 	g_autofree gchar *email = NULL;
 	g_autofree gchar *username = NULL;
 	g_autofree gchar *target_repo = NULL;
+	FILE *file;
 	gchar *tmp;
 	gchar *tmp2;
 	gchar *tmp3;
-
-	gpgme_data_t sigdata;
-	#define BUF_SIZE 512
-	char buf[BUF_SIZE + 1];
-	FILE *file;
-	gint ret;
-	GError *tmp_error = NULL;
+	LiPkgBuilderPrivate *priv = GET_PRIVATE (builder);
 
 	cdata = li_config_data_new ();
 
@@ -643,35 +670,60 @@ li_pkg_builder_write_dsc_file (LiPkgBuilder *builder, const gchar *pkg_fname_rt,
 	/* set target repository */
 	li_config_data_set_value (cdata, "Target", target_repo);
 
-	tmp = li_config_data_get_data (cdata);
-	sigdata = li_pkg_builder_sign_data (builder, tmp, GPGME_SIG_MODE_CLEAR, &tmp_error);
-	g_free (tmp);
-	if (tmp_error != NULL) {
-		g_propagate_error (error, tmp_error);
-		return;
-	}
-	g_print ("%s\n", _("DSC file signed."));
-
 	fname = g_strdup_printf ("%s.dsc", pkg_fname_rt);
-	tmp = li_config_data_get_data (cdata);
+	if (priv->sign_package) {
+		gpgme_data_t sigdata;
+		#define BUF_SIZE 512
+		char buf[BUF_SIZE + 1];
+		gint ret;
+		GError *tmp_error = NULL;
+
+		tmp = li_config_data_get_data (cdata);
+		sigdata = li_pkg_builder_sign_data (builder, tmp, GPGME_SIG_MODE_CLEAR, &tmp_error);
+		g_free (tmp);
+		if (tmp_error != NULL) {
+			g_propagate_error (error, tmp_error);
+			return;
+		}
+		g_print ("%s\n", _("DSC file signed."));
+
+		file = fopen (fname, "w");
+		if (file == NULL) {
+			g_set_error (error,
+				LI_BUILDER_ERROR,
+				LI_BUILDER_ERROR_SIGN,
+				_("Unable to write signed DSC file: %s"),
+				g_strerror (errno));
+			gpgme_data_release (sigdata);
+			return;
+		}
+
+		gpgme_data_seek (sigdata, 0, SEEK_SET);
+		while ((ret = gpgme_data_read (sigdata, buf, BUF_SIZE)) > 0)
+			fwrite (buf, sizeof (char), ret, file);
+		fclose (file);
 
-	file = fopen (fname, "w");
-	if (file == NULL) {
-		g_set_error (error,
-			LI_BUILDER_ERROR,
-			LI_BUILDER_ERROR_SIGN,
-			_("Unable to write signature on dsc file: %s"),
-			g_strerror (errno));
 		gpgme_data_release (sigdata);
-		return;
+	} else {
+		/* Just write file without signature */
+		g_print ("%s\n", _("DSC file will not be signed."));
+
+		file = fopen (fname, "w");
+		if (file == NULL) {
+			g_set_error (error,
+				LI_BUILDER_ERROR,
+				LI_BUILDER_ERROR_SIGN,
+				_("Unable to write DSC file: %s"),
+				g_strerror (errno));
+			return;
+		}
+
+		tmp = li_config_data_get_data (cdata);
+		fwrite (tmp, sizeof (char), strlen (tmp), file);
+		g_free (tmp);
+
+		fclose (file);
 	}
-
-	gpgme_data_seek (sigdata, 0, SEEK_SET);
-	while ((ret = gpgme_data_read (sigdata, buf, BUF_SIZE)) > 0)
-		fwrite (buf, ret, 1, file);
-	fclose (file);
-
-	gpgme_data_release (sigdata);
 }
 
 /**
@@ -817,6 +869,7 @@ li_pkg_builder_build_package_with_details (LiPkgBuilder *builder, LiPkgInfo *ctl
 
 	if (priv->sign_package) {
 		sig_fname = li_pkg_builder_sign_package (builder, tmp_dir, sign_files, &tmp_error);
+		g_print ("Package for '%s' signed.\n", as_component_get_id (cpt));
 		if (tmp_error != NULL) {
 			g_propagate_error (error, tmp_error);
 			return FALSE;
@@ -900,20 +953,19 @@ li_pkg_builder_create_package_from_dir (LiPkgBuilder *builder, const gchar *dir,
 	g_object_ref (cpt);
 
 	if (out_fname == NULL) {
-		gchar *tmp;
+		g_autofree gchar *pkgname = NULL;
 		const gchar *version;
 
 		/* we need to auto-generate a package filename */
-		tmp = li_str_replace (as_component_get_name (cpt), " ", "");
+		pkgname = li_get_pkgname_from_component (cpt);
 		version = li_get_last_version_from_component (cpt);
 		if (version != NULL) {
-			pkg_fname_rt  = g_strdup_printf ("%s/%s-%s.ipk", dir, tmp, version);
-			pkg_fname_sdk = g_strdup_printf ("%s/%s-%s.devel.ipk", dir, tmp, version);
+			pkg_fname_rt  = g_strdup_printf ("%s/%s-%s.ipk", dir, pkgname, version);
+			pkg_fname_sdk = g_strdup_printf ("%s/%s-%s.devel.ipk", dir, pkgname, version);
 		} else {
-			pkg_fname_rt  = g_strdup_printf ("%s/%s.ipk", dir, tmp);
-			pkg_fname_sdk = g_strdup_printf ("%s/%s.devel.ipk", dir, tmp);
+			pkg_fname_rt  = g_strdup_printf ("%s/%s.ipk", dir, pkgname);
+			pkg_fname_sdk = g_strdup_printf ("%s/%s.devel.ipk", dir, pkgname);
 		}
-		g_free (tmp);
 	} else {
 		g_autofree gchar *base_fname;
 		g_autofree gchar *dirname;
