@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
  *
- * Copyright (C) 2015 Matthias Klumpp <matthias@tenstral.net>
+ * Copyright (C) 2015-2016 Matthias Klumpp <matthias@tenstral.net>
  *
  * Licensed under the GNU Lesser General Public License Version 2.1
  *
@@ -38,6 +38,8 @@
 #include <sys/wait.h>
 #include <sys/mount.h>
 #include <sched.h>
+#include <grp.h>
+#include <pwd.h>
 
 #include "li-utils.h"
 #include "li-utils-private.h"
@@ -112,23 +114,24 @@ li_build_master_init (LiBuildMaster *bmaster)
 	priv->build_root = NULL;
 	priv->dep_data_paths = NULL;
 	priv->get_shell = FALSE;
+
+	priv->build_uid = getuid ();
+	priv->build_gid = getgid ();
 }
 
 /**
  * li_build_master_check_dependencies:
  */
 static void
-li_build_master_check_dependencies (LiPackageGraph *pg, LiManager *mgr, GNode *root, GError **error)
+li_build_master_check_dependencies (LiPackageGraph *pg, LiManager *mgr, LiPkgInfo *pki, gboolean use_builddeps, GError **error)
 {
-	g_autoptr(GList) all_pkgs = NULL;
+	g_autoptr(GPtrArray) all_pkgs = NULL;
 	GError *tmp_error = NULL;
-	LiPkgInfo *pki;
 	g_autoptr(GPtrArray) deps = NULL;
+	g_autoptr(GPtrArray) install_todo = NULL;
 	guint i;
 
-	pki = LI_PKG_INFO (root->data);
-
-	if (G_NODE_IS_ROOT (root)) {
+	if (use_builddeps) {
 		/* we need to take the build-deps from the package we want to build... */
 		deps = li_parse_dependencies_string (li_pkg_info_get_build_dependencies (pki));
 	} else {
@@ -146,6 +149,7 @@ li_build_master_check_dependencies (LiPackageGraph *pg, LiManager *mgr, GNode *r
 		return;
 	}
 
+	install_todo = g_ptr_array_new ();
 	for (i = 0; i < deps->len; i++) {
 		LiPkgInfo *ipki;
 		gboolean ret;
@@ -171,26 +175,39 @@ li_build_master_check_dependencies (LiPackageGraph *pg, LiManager *mgr, GNode *r
 					_("Could not find bundle '%s' which is necessary to build this software."),
 					li_pkg_info_get_name (dep));
 		} else if (li_pkg_info_has_flag (ipki, LI_PACKAGE_FLAG_INSTALLED)) {
-			GNode *node;
-
 			/* dependency is already installed, add it as satisfied */
-			node = li_package_graph_add_package (pg, root, ipki, dep);
+			li_package_graph_add_package (pg, pki, ipki, dep);
 
 			/* we need a full dependency tree */
-			li_build_master_check_dependencies (pg, mgr, node, &tmp_error);
+			li_build_master_check_dependencies (pg, mgr, ipki, FALSE, &tmp_error);
 			if (tmp_error != NULL) {
 				g_propagate_error (error, tmp_error);
 				return;
 			}
 		} else {
-			g_set_error (error,
-					LI_BUILD_MASTER_ERROR,
-					LI_BUILD_MASTER_ERROR_BUILD_DEP_MISSING,
-					_("Bundle '%s' needs to be installed in order to build this software."),
-					li_pkg_info_get_id (dep));
-			return;
+			g_ptr_array_add (install_todo, dep);
+			continue;
 		}
 	}
+
+	if (install_todo->len > 0) {
+		g_autoptr(GString) depline = NULL;
+
+		depline = g_string_new ("");
+		for (i = 0; i < install_todo->len; i++) {
+			LiPkgInfo *dep = LI_PKG_INFO (g_ptr_array_index (install_todo, i));
+			g_string_append_printf (depline, "%s ", li_pkg_info_get_id (dep));
+		}
+		if (depline->len > 0)
+			g_string_truncate (depline, depline->len - 1);
+
+		g_set_error (error,
+				LI_BUILD_MASTER_ERROR,
+				LI_BUILD_MASTER_ERROR_BUILD_DEP_MISSING,
+				_("Bundle(s) '%s' need to be installed in order to build this software."),
+				depline->str);
+		return;
+	}
 }
 
 /**
@@ -204,7 +221,6 @@ li_build_master_resolve_builddeps (LiBuildMaster *bmaster, LiPkgInfo *pki, GErro
 	g_autoptr(GHashTable) depdirs = NULL;
 	g_autoptr(LiManager) mgr = NULL;
 	guint i;
-	GNode *root;
 	GError *tmp_error = NULL;
 	LiBuildMasterPrivate *priv = GET_PRIVATE (bmaster);
 
@@ -217,17 +233,16 @@ li_build_master_resolve_builddeps (LiBuildMaster *bmaster, LiPkgInfo *pki, GErro
 		return;
 	}
 
-	li_package_graph_set_root (pg, pki);
-	root = li_package_graph_get_root (pg);
 	mgr = li_manager_new ();
+	li_package_graph_add_package (pg, NULL, pki, NULL);
 
-	li_build_master_check_dependencies (pg, mgr, root, &tmp_error);
+	li_build_master_check_dependencies (pg, mgr, pki, TRUE, &tmp_error);
 	if (tmp_error != NULL) {
 		g_propagate_error (error, tmp_error);
 		return;
 	}
 
-	full_deps = li_package_graph_branch_to_array (root);
+	full_deps = li_package_graph_branch_to_array (pg, pki, FALSE);
 	if (full_deps == NULL) {
 		g_warning ("Building package with no build-dependencies defined.");
 		return;
@@ -345,6 +360,7 @@ li_build_master_print_section (LiBuildMaster *bmaster, const gchar *section_name
 	g_string_free (str, TRUE);
 }
 
+#if 0
 /**
  * li_build_master_exec:
  */
@@ -360,6 +376,83 @@ li_build_master_exec (LiBuildMaster *bmaster, const gchar *cmd)
 
 	return res;
 }
+#endif
+
+/**
+ * li_build_master_exec_command_sequence:
+ *
+ */
+static gint
+li_build_master_exec_command_sequence (LiBuildMaster *bmaster, const gchar *stage_id, GPtrArray *cmds, const gchar *env_fname)
+{
+	gint fd;
+	guint i;
+	g_autofree gchar *tmp_fname = NULL;
+	g_autofree gchar *scmd = NULL;
+	gint res = 0;
+
+	/* The system() command always spawns a new shell, so if you set
+	 * environment vars or change directories in build.yml it won't work.
+	 * So we cheat and execute a shell script instead, which even sources
+	 * the environment from previous steps.
+	 * This is more hack than a proper solution, so if you have a better idea
+	 * for this, please implement it. */
+
+	tmp_fname = g_strdup_printf ("/tmp/%s-XXXXXX", stage_id);
+	fd = mkstemp (tmp_fname);
+	g_debug ("Command script: %s", tmp_fname);
+	if (fd < 0) {
+		g_error ("Unable to store script for %s.", stage_id);
+		return 1;
+	}
+
+	res = g_chmod (tmp_fname, 0775);
+	if (res < 0) {
+		g_error ("Unable to set permissions: %s", g_strerror (errno));
+		return res;
+	}
+
+	if (dprintf (fd, "#!/bin/sh\n. %s\nset -e\n\n", env_fname) < 0) {
+		g_error ("Unable to write command sequence.");
+		return 1;
+	}
+
+	for (i = 0; i < cmds->len; i++) {
+		gchar *cmd;
+		gchar *tmp;
+		g_autofree gchar *msg = NULL;
+		cmd = (gchar*) g_ptr_array_index (cmds, i);
+
+		tmp = g_shell_quote (cmd);
+		msg = li_str_replace (tmp, "'", "");
+		g_free (tmp);
+
+		if (dprintf (fd, "echo ' ! %s'\n", msg) < 0) {
+			g_error ("Unable to write command sequence.");
+			return 1;
+		}
+		if (dprintf (fd, "%s\n", cmd) < 0) {
+			g_error ("Unable to write command sequence.");
+			return 1;
+		}
+		dprintf (fd, "\n");
+	}
+
+	/* ensure we export the environment to our environment file */
+	if (dprintf (fd, "export > %s\n", env_fname) < 0) {
+		g_error ("Unable to write command sequence.");
+		return 1;
+	}
+
+	/* run command script */
+	scmd = g_strdup_printf ("sh %s", tmp_fname);
+	res = system (scmd);
+	if (res > 255)
+		res = 1;
+	close (fd);
+
+	return res;
+}
 
 /**
  * li_build_master_mount_deps:
@@ -367,12 +460,15 @@ li_build_master_exec (LiBuildMaster *bmaster, const gchar *cmd)
  * Mount the depdenencies into the environment as an overlay.
  */
 static gint
-li_build_master_mount_deps (LiBuildMaster *bmaster, const gchar *chroot_dir)
+li_build_master_mount_deps (LiBuildMaster *bmaster, const gchar *chroot_dir, const gchar *env_root)
 {
 	guint i;
 	gint res;
 	g_autoptr(GString) lowerdirs = NULL;
 	g_autofree gchar *mount_target = NULL;
+	g_autofree gchar *volatile_data_dir = NULL;
+	g_autofree gchar *ofs_wdir = NULL;
+	g_autofree gchar *app_mount_target = NULL;
 	gchar *tmp;
 	LiBuildMasterPrivate *priv = GET_PRIVATE (bmaster);
 
@@ -387,16 +483,18 @@ li_build_master_mount_deps (LiBuildMaster *bmaster, const gchar *chroot_dir)
 	mount_target = g_build_filename (chroot_dir, "usr", NULL);
 
 	g_debug ("Mounting build dependencies into environment.");
-	lowerdirs = g_string_new (mount_target);
+	lowerdirs = g_string_new ("");
 	for (i = 0; priv->dep_data_paths[i] != NULL; i++) {
 		const gchar *dep_data_path = priv->dep_data_paths[i];
-
-		g_string_append_printf (lowerdirs, ":%s", dep_data_path);
+		g_string_append_printf (lowerdirs, "%s:", dep_data_path);
 	}
+	g_string_append (lowerdirs, mount_target);
 
+	/* IMPORTANT: We do *not* mount the root filesystem NOSUID, so the build process can use sudo on demand.
+	 * All other stuff and especially the environment when running the app *must not* perform mounts with SUID allowed. */
 	tmp = g_strdup_printf ("lowerdir=%s", lowerdirs->str);
 	res = mount ("overlay", mount_target,
-				 "overlay", MS_MGC_VAL | MS_RDONLY | MS_NOSUID, tmp);
+				 "overlay", MS_MGC_VAL | MS_RDONLY, tmp);
 	g_free (tmp);
 	if (res != 0) {
 		fprintf (stderr, "Unable to mount directory. %s\n", strerror (errno));
@@ -404,13 +502,42 @@ li_build_master_mount_deps (LiBuildMaster *bmaster, const gchar *chroot_dir)
 		goto out;
 	}
 
-	/* bind-mount /app to /usr, so binaries compiled with that prefix can find their data  */
-	tmp = g_build_filename (chroot_dir, LI_SW_ROOT_PREFIX, NULL);
-	res = mount (mount_target, tmp,
-				 NULL, MS_BIND, NULL);
+	/* Now we mount /app as an overlay on /usr and make it writable using OverlayFS.
+	 * That way, binaries compiled with that prefix can find their data and we allow
+	 * the setup process to place data here which can then be found by other steps in
+	 * the same stup process. This is useful for creating one bundle out of multiple
+	 * submodules. */
+
+	/* create volatile data dir for /app */
+	volatile_data_dir = g_build_filename (env_root, "volatile_app", NULL);
+	res = g_mkdir_with_parents (volatile_data_dir, 0755);
+	if (res != 0) {
+		g_warning ("Unable to set up the environment (volatile /app): %s", g_strerror (errno));
+		goto out;
+	}
+
+	/* create OverlayFS work dir */
+	ofs_wdir = g_build_filename (env_root, "ofs_work_app", NULL);
+	res = g_mkdir_with_parents (ofs_wdir, 0755);
+	if (res != 0) {
+		g_warning ("Unable to set up the environment (workdir /app): %s", g_strerror (errno));
+		goto out;
+	}
+
+	/* now mount our build-data directory via OverlayFS */
+	tmp = g_strdup_printf ("lowerdir=%s,upperdir=%s,workdir=%s", mount_target, volatile_data_dir, ofs_wdir);
+	app_mount_target = g_build_filename (chroot_dir, LI_SW_ROOT_PREFIX, NULL);
+	res = mount ("overlay", app_mount_target,
+			 "overlay", MS_MGC_VAL | MS_NOSUID, tmp);
 	g_free (tmp);
 	if (res != 0) {
-		g_warning ("Unable to set up the environment: %s", g_strerror (errno));
+		g_warning ("Unable to set up the environment (/app mount): %s", g_strerror (errno));
+		goto out;
+	}
+
+	res = chown (volatile_data_dir, priv->build_uid, priv->build_gid);
+	if (res != 0) {
+		g_warning ("Could not adjust permissions on volatile /app dir: %s", g_strerror (errno));
 		goto out;
 	}
 
@@ -418,6 +545,8 @@ out:
 	if (res != 0) {
 		if (mount_target != NULL)
 			umount (mount_target);
+		if (app_mount_target != NULL)
+			umount (app_mount_target);
 	}
 
 	return res;
@@ -434,12 +563,13 @@ li_build_master_run_executor (LiBuildMaster *bmaster, const gchar *env_root)
 {
 	gint res = 0;
 	gboolean ret;
-	guint i;
 	gchar *tmp;
 	g_autofree gchar *build_data_root = NULL;
 	g_autofree gchar *newroot_dir = NULL;
 	g_autofree gchar *volatile_data_dir = NULL;
 	g_autofree gchar *ofs_wdir = NULL;
+	g_autofree gchar *env_tmp_fname = NULL;
+	gint env_fd = 0;
 	LiBuildMasterPrivate *priv = GET_PRIVATE (bmaster);
 
 	newroot_dir = li_run_env_setup_with_root (priv->chroot_orig_dir);
@@ -483,7 +613,7 @@ li_build_master_run_executor (LiBuildMaster *bmaster, const gchar *env_root)
 	}
 
 	/* overlay the base filesystem with the build-dependency data */
-	res = li_build_master_mount_deps (bmaster, newroot_dir);
+	res = li_build_master_mount_deps (bmaster, newroot_dir, env_root);
 	if (res != 0) {
 		g_warning ("Unable to set up the environment: %s", g_strerror (errno));
 		goto out;
@@ -505,11 +635,26 @@ li_build_master_run_executor (LiBuildMaster *bmaster, const gchar *env_root)
 		goto out;
 	}
 
-	/* we now finished everything we needed root for, so drop root in case we build as user */
-	if (chown (volatile_data_dir, priv->build_uid, priv->build_gid) != 0) {
+	/* ensure we can write volatile data */
+	res = chown (volatile_data_dir, priv->build_uid, priv->build_gid);
+	if (res != 0) {
 		g_warning ("Could not adjust permissions on volatile data dir: %s", g_strerror (errno));
 		goto out;
 	}
+
+	/* try to initialize groups, failure is not fatal */
+	if (priv->build_uid > 0) {
+		struct passwd *upws;
+
+		upws = getpwuid (priv->build_uid);
+		if (upws != NULL) {
+			if (initgroups (upws->pw_name, priv->build_gid) < 0)
+				g_warning ("Unable to initialize user groups: %s", g_strerror (errno));
+		} else {
+			g_warning ("Unable to initialize user groups: Could not determine user name: %s", g_strerror (errno));
+		}
+	}
+	/* we now finished everything we needed root for, so drop root in case we build as user */
 	if (setgid (priv->build_gid) != 0) {
 		g_warning ("Unable to set gid: %s", g_strerror (errno));
 		goto out;
@@ -524,41 +669,56 @@ li_build_master_run_executor (LiBuildMaster *bmaster, const gchar *env_root)
 				 priv->email,
 				 priv->target_repo);
 
+	/* prepare our environment-file hack */
+	env_tmp_fname = g_strdup ("/tmp/environment-XXXXXX");
+	env_fd = mkstemp (env_tmp_fname);
+	if (env_fd < 0) {
+		g_error ("Unable to create environment file: %s", g_strerror (errno));
+		res = 1;
+		goto out;
+	}
+	res = g_chmod (env_tmp_fname, 0775);
+	if (res < 0) {
+		g_error ("Unable to set permissions: %s", g_strerror (errno));
+		goto out;
+	}
+
+	/* now start running the actual commands */
 	li_build_master_print_section (bmaster, "Preparing Build Environment");
 	if (priv->cmds_pre != NULL) {
-		for (i = 0; i < priv->cmds_pre->len; i++) {
-			gchar *cmd;
-			cmd = (gchar*) g_ptr_array_index (priv->cmds_pre, i);
-			res = li_build_master_exec (bmaster, cmd);
-			if (res != 0)
-				goto out;
-		}
+		res = li_build_master_exec_command_sequence (bmaster,
+								"prepare",
+								priv->cmds_pre,
+								env_tmp_fname);
+		if (res != 0)
+			goto out;
 	}
 
 	if (priv->get_shell) {
+		g_autofree gchar *cmd = NULL;
+
 		g_debug ("Starting new shell session...");
-		system ("sh");
+		cmd = g_strdup_printf ("sh -sc '. %s'", env_tmp_fname);
+		system (cmd);
 	} else {
 		/* we don't start an interactive shell, and get to business instead */
 		li_build_master_print_section (bmaster, "Build");
-		for (i = 0; i < priv->cmds->len; i++) {
-			gchar *cmd;
-			cmd = (gchar*) g_ptr_array_index (priv->cmds, i);
-			res = li_build_master_exec (bmaster, cmd);
-			if (res != 0)
-				goto out;
-		}
+		res = li_build_master_exec_command_sequence (bmaster,
+								"build",
+								priv->cmds,
+								env_tmp_fname);
+		if (res != 0)
+			goto out;
 	}
 
 	li_build_master_print_section (bmaster, "Cleanup");
 	if (priv->cmds_post != NULL) {
-		for (i = 0; i < priv->cmds_post->len; i++) {
-			gchar *cmd;
-			cmd = (gchar*) g_ptr_array_index (priv->cmds_post, i);
-			res = li_build_master_exec (bmaster, cmd);
-			if (res != 0)
-				goto out;
-		}
+		res = li_build_master_exec_command_sequence (bmaster,
+								"cleanup",
+								priv->cmds_post,
+								env_tmp_fname);
+		if (res != 0)
+			goto out;
 	}
 
 out:
@@ -567,6 +727,8 @@ out:
 		umount (tmp);
 		g_free (tmp);
 	}
+	if (env_fd > 0)
+		close (env_fd);
 
 	return res;
 }
@@ -621,16 +783,16 @@ li_build_master_run (LiBuildMaster *bmaster, GError **error)
 
 	/* fork our build helper */
 	pid = fork ();
-
 	if (pid == 0) {
 		/* child process */
+
 		res = li_build_master_run_executor (bmaster, env_root);
 		exit (res);
 	} else if (pid < 0) {
 		/* error */
 		g_set_error_literal (error,
 				LI_BUILD_MASTER_ERROR,
-				LI_BUILD_MASTER_ERROR_FAILED,
+				LI_BUILD_MASTER_ERROR_INIT,
 				"Unable to fork.");
 	}
 
@@ -659,7 +821,8 @@ li_build_master_run (LiBuildMaster *bmaster, GError **error)
 
 	g_debug ("Executor is done, rescuing build artifacts...");
 	tmp = g_build_filename (env_root, "volatile", "lipkg", NULL);
-	artifacts = li_utils_find_files_matching (tmp, "*.ipk*", FALSE);
+	if (g_file_test (tmp, G_FILE_TEST_IS_DIR))
+		artifacts = li_utils_find_files_matching (tmp, "*.ipk*", FALSE);
 	g_free (tmp);
 	if ((artifacts == NULL) || (artifacts->len == 0)) {
 		g_print ("Unable to find build artifacts!\n");
@@ -699,7 +862,7 @@ finish:
 	li_delete_dir_recursive (env_root);
 
 out:
-	if (res != 0)
+	if ((res != 0) && (*error == NULL))
 		g_set_error_literal (error,
 				LI_BUILD_MASTER_ERROR,
 				LI_BUILD_MASTER_ERROR_STEP_FAILED,
